/*
	Copyright (c) Frank Klemm 1992. All Rights reserved.
	Frank.Klemm@HpUx.Rz.Uni-Jena.De

	Ported to linux 5 Feb 93 
	  Ray Balister - balister@cebaf.gov

	Ported to NetBSD-0.9 on 29 Oct 93
	  also gathered into one file.
	  Daisuke Yokota - yokota@tom.comm.waseda.ac.jp
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<unistd.h>

#ifdef Linux
#include <asm/system.h>
#endif /*Linux*/

#ifdef NetBSD
#include	<sys/file.h>
#include	<sys/ioctl.h>
#include	<machine/cpufunc.h>
#endif /*NetBSD*/

#define TRUE      1
#define FALSE     0

/**************************************************************************/
/* dos.h of original distribution */

#ifdef Linux

inline void outportb( unsigned short port, char value) {
	__asm__ __volatile__ ("outb %%al,%%dx"
		::"a" ((char) value),"d" ((unsigned short) port));
}

inline unsigned int inportb(unsigned short port) {
	unsigned int _v;
	__asm__ __volatile__ ("inb %%dx,%%al"
		:"=a" (_v):"d" ((unsigned short) port),"0" (0));
	return _v;
}

#endif /*Linux*/
/**************************************************************************/


/**************************************************************************/
/* perm_ports (for Linux) */

/*
** Routine to set the permissions of i/o
** ports used for clock manipulation.  May not
** be clean, but it works.  --  RLB
*/

#ifdef Linux

void perm_ports(int perm) {
	ioperm(0x3c2, 1, perm);
	ioperm(0x3cc, 1, perm);
}

#endif /*Linux*/
/**************************************************************************/


/**************************************************************************/
/* vga_open, vga_close (for NetBSD-0.9) */

#ifdef NetBSD

#define CONSOLE_DEV		"/dev/vga"
#define	CONSOLE_X_MODE_ON	_IO('t',121) /* I've got from               */
#define	CONSOLE_X_MODE_OFF	_IO('t',122) /*  sys/arch/i386/isa/pccons.c */

#define inportb  inb
#define outportb outb

static	int	fd_vga; /* file desriptor of CONSOLE_DEV */

/* open CONSOLE_DEV and get the file descriptor */
void vga_open() {
	printf("Opening %s on NetBSD.\n", CONSOLE_DEV);
	if ( (fd_vga = open(CONSOLE_DEV,O_RDWR)) == -1 ) {
		perror("open");
		exit(-1);
	}
	return;
}

/* get the I/O permission */
void vga_perm_get() {
	printf("Getting I/O permission on NetBSD.\n");
	if ( ioctl(fd_vga, CONSOLE_X_MODE_ON) == -1 ) {
		perror("ioctl");
		exit(-1);
	}
	return;
}

/* clear the I/O permission */
void vga_perm_clear() {
	printf("Clearing I/O permission on NetBSD.\n");
	if ( ioctl(fd_vga,CONSOLE_X_MODE_OFF) == -1 ) {
		perror("ioctl");
		exit(-1);
	}
}

/* close CONSOLE_DEV */
void vga_close() {
	printf("Closing %s on NetBSD.\n", CONSOLE_DEV);
	close(fd_vga);
	return;
}

#endif /*NetBSD*/
/**************************************************************************/


/**************************************************************************/
/* setclock.c of original distribution */

/* Romeved 29 Oct 93 - D.Y
extern inline void outportb(char value, unsigned short port);
extern inline unsigned int inportb(unsigned short port);

#ifndef enable
#define enable()        sti()
#endif

#ifndef disable
#define disable()       cli()
#endif
*/

void setdotclock(void *ptr) {
	register unsigned char a;
	register unsigned i;
	unsigned long m;

	a=inportb(0x3CC) & ~0x0C;

#define S(x)	outportb(0x3C2,a | 4*(x))

/*    disable();    Removed 5 Feb 93 - RLB*/
	for (i=0; i<5; i++)
		S(2), S(3);
	for (i=0; i<2; i++)
		S(0), S(1);
	for (i=0, m=1; i<24; i++, m+=m)
	if (*(unsigned long *)ptr & m)
		S(1), S(0), S(2), S(3);
	else
		S(3), S(2), S(0), S(1);
	S(3), S(2), S(3);
	S(3);
/*    enable();   Removed 5 Feb 93 - RLB*/

#undef S
}
/* End of setclock.c */
/**************************************************************************/

/**************************************************************************/
/* setfreq.c of original distribution */

/* Removed 29 Oct 93 - D.Y
extern inline void outportb(char value, unsigned short port);
extern inline unsigned int inportb(unsigned short port);
*/

unsigned _bin, _mul, _div, _gen;
unsigned long _value;

/*
	This function use integer arithmetic instead of floating point.
*/

unsigned long setfreq(unsigned long freq, unsigned quarz) {
	register unsigned i;
	unsigned _err=32767;

#define MUL		1000
#define QUARZ		(2*14318lu)
#define QUARZMUL	(MUL*QUARZ)

	/* Determines binary divider _bin */
	_bin=0;
	while (freq<40000lu && _bin<7)
		freq <<= 1, _bin++;

	/* Determines _mul and _div */
	for (i=0; i<128; i++) {
		register unsigned long t1=QUARZMUL*(i+3)/freq;
		register unsigned long t2=(t1+MUL/2)/MUL;
		register unsigned long t3=t1-t2*MUL;
		t2 -= 2;
		t3=(long)t3 > 0 ? t3 : -t3;
		t3=t3*127/(i+3);
		if (t2 > 127)
			continue;
		if (t3 < _err)
			_err=(unsigned)t3, _mul=i, _div=(unsigned)t2;
	}

	/* Determines the range variable _gen */
	if (freq<40000l)
		_gen=0;
	else if (freq<50000l)
		_gen=1;
	else if (freq<72000l)
		_gen=3;
	else
		_gen=7;

	/* Calculates the value to set */
	_value=((unsigned long)quarz<<21) |
		((unsigned long)_gen <<17) |
		((unsigned long)_mul <<10) | (_bin<<7) | _div;

	/* And set it */
	setdotclock(&_value);
	outportb(0x3C2,(inportb(0x3CC) & 0xF3) | ((quarz << 2) & 0x0C));

	/* returns the real frequency set */
	return (((2*QUARZ*(_mul+3)/(_div+2))>>_bin)+1)>>1;

#undef MUL
#undef QUARZ
#undef QUARZMUL

}
/* End of setfreq.c */
/**************************************************************************/


/**************************************************************************/
/* freq.c of original distribution                  */
/* modified for NetBSD by D.Yokota                  */
/*   added NetBSD I/O permission call,              */
/*   changed variable name from "quarz" to "quartz" */

int main(int argc, char **argv) {
	unsigned long freq;
	unsigned long newfreq;
	unsigned quartz=2;

	if (argc == 1) {
		fprintf(stderr,"usage: FREQ Pixel_clock_in_kHz [Quartz_#]\n");
		return 1;
	}

	freq=atol(argv[1]);
	if (argc > 2)
		quartz=atoi(argv[2]);

	if (freq<300l || freq>128000l) {
		fprintf(stderr,"Pixel clock must be in the range [300-128000]\n");
		return 1;
	}
	if (freq>86000)
		fprintf(stderr,"Clock generator is running outside the specification\n");
	if (quartz>2) {
		fprintf(stderr,"Quartz must be 0, 1, or 2\n");
		return 1;
	}

#ifdef Linux
	perm_ports( TRUE );  /* set i/o permissions - RLB */
#endif

#ifdef NetBSD
	vga_open(); /* open vga (NetBSD) - D.Y */
	vga_perm_get(); /* get i/o permissions (NetBSD) - D.Y */
#endif

	newfreq=setfreq(freq,quartz);
	printf("[%1X-%1X-%2X-%2X]=[%06lX]\t%5lu kHz instead of%6lu kHz\n",
					_bin,_gen,_mul,_div,_value,newfreq,freq);

#ifdef Linux
	perm_ports( FALSE ); /* clear i/o permissions - RLB */
#endif /*Linux*/

#ifdef NetBSD
	if (quartz == 0) /* Clearing only when leaving X, otherwise hangs up. */
		vga_perm_clear(); /* clear i/o permissions (NetBSD) - D.Y */
	vga_close(); /* close vga (NetBSD) - D.Y */
#endif
	return 0;
}
/* End of freq.c */
/**************************************************************************/
